// Keep in mind that this only represents a single possible solution.
// When analyzing a problem, the experience of the designer will tend
// to influence the resulting design.

#include <list>
using namespace std;

// Use typedefs to hide the implementation of the data type.  This
// allows us to more easily change "Money" into a class of its own
// in later stages of the development.
typedef long BookId;
typedef long PatronId;

typedef float Money;		// Placeholder
typedef int PaymentType;	//   (ditto)

// This is a "given".  It's only here to illustrate a point.
template <class T, class I = int>
class ListOf : public list<T> {
    // ...
  public:
    void Add(T value);
    void Del(T value);
    T First();		// Should be iterators, but we don't
    T Last();		//   learn those until Chapter 4 :)
    T Next();
    T Prev();
    T Find(I &value);	// Given BookId, return Book
};

class Date {
    static long today;
    long thisDate;
  public:
    static Date &getToday();
    bool isValid() { return thisDate != -1; }

    friend int operator-(Date&, Date&);
    friend Date &operator+(int,   Date&);
    friend Date &operator+(Date&, int);
};

// Classes which use CheckOutPolicy may want to refer to
// a given instance of this class, or may want to copy the
// information into themselves.  This is an implementation
// issue and isn't important as far as the design is
// concerned.
class CheckOutPolicy {
    int daysBeforeDue;		// Referenced by CheckOut
    Money penniesPerDay;	// Referenced by Fine
  public:
    int getDaysBeforeDue() { return daysBeforeDue; }
    // ...
};
ListOf<CheckOutPolicy*>	coPolicies;	// all CheckOutPolicys

// Each patron object refers to one of these.  Most patrons
// share the same LimitPolicy, but some patrons, such as
// staff, will likely have a more lenient policy.
class LimitPolicy {
    int bookLimit;		// max number of books at once
    // Other per-Patron policies?
  public:
    // ...
};
ListOf<LimitPolicy*>	lPolicies;

// Remember: a more generic interface would have "Media" as a
// base class, with Book, LP, Video, and so on, as subclasses.
class BookInfo {
    char *author;
    char *title;
    char *publisher;
    char *isbn;
  public:
    // ...
};
class Book {
    BookId    bookId;		// One per physical book
    BookInfo &info;		// All books with the same info
    CheckOutPolicy &cop;	// Business policy...
    Money     cost;		// Fines shouldn't be > cost!
  public:
    // Constructors and destructors will be needed...
    CheckOutPolicy *getCheckOutPolicy();
};
ListOf<Book*, BookId>	catalog;	// All books
ListOf<Book*, BookId>	lostBooks;	// Or a flag in Book?

class Patron {
    PatronId custId;		// customer id, so to speak
    char *name, *addr, *city;	// personal information
    // This next field specifies an alternate policy for those
    // patrons who might receive a different set of limits than
    // that imposed by the book itself.  For example, staff often
    // have limits that are different than ordinary Book-specified
    // values.
    CheckOutPolicy *cop;	// Overrides 'Book::cop'
    LimitPolicy &lp;
  public:
    // Constructors and destructors will be needed...
    CheckOutPolicy *getCheckOutPolicy();
};
ListOf<Patron*, PatronId>	customers;

// Need to keep a list of these in order to record payments.
class Payment {
    Date     whenPaid;
    Money    amtPaid;
    PatronId librarian;		// Librarian who accepted payment
    PaymentType type;		// Credit, cash, check, debit card.
  public:
    Money getAmountPaid() { return amtPaid; }
};

// Created every time a book is checked out.  The list
// of all CheckOuts is scanned once per day and over-due
// books create "Fine" objects and link back to the CheckOut.
// Note that a Fine is NOT a CheckOut, but does contain a
// reference to the original CheckOut object.
class CheckOut {
    PatronId custId;		// could be Patron*
    BookId   bookId;		// could be Book*
    Date     dueBack;		// copied policy
  public:
    CheckOut(BookId b, PatronId p) {
	bookId = b;
	custId = p;
	dueBack =
	    catalog.Find(b)->
	    getCheckOutPolicy()->
	    getDaysBeforeDue() + Date::getToday();
    }
    Date &getDueDate() { return dueBack; }
};
// Entries are always deleted from "checkOuts" when returned.  Also,
// they are delete'd to free up the memory, unless they are in use
// in a Fine.  In that case, the Fine is responsible for delete'ing
// them in its own destructor.
ListOf<CheckOut*>	checkOuts;

class Fine {
    CheckOut *co;		// To the original CheckOut
    Money penniesPerDay;	// copied from policy (speed)
    Date  whenReturned;		// actual return date
    ListOf<Payment*> pymts;
  public:
    Fine(const CheckOut &co);
    Money AmountOwed() {
	Date date;
	Money payments = 0.0;
	ListOf<Payment*>::iterator i = pymts.begin();
	while (i != pymts.end())
	    payments += (*i++)->getAmountPaid();
	if (whenReturned.isValid())	date = whenReturned;
	else				date = Date::getToday();
	return penniesPerDay * (date - co->getDueDate()) - payments;
    }
};
// If CheckOuts become overdue, they become "Fines".
ListOf<Fine*>     	fines;
